import unittest
import numpy as np
from numpy.testing import assert_allclose
from problems import integrator_control_problem, bang_bang_problem
from casadi import vertcat, DM, hcat
from rockit import MultipleShooting, DirectCollocation, Ocp, SingleShooting, SplineMethod

class OcpSolutionTests(unittest.TestCase):
    def test_grid_integrator(self):
        N, T, u_max, x0 = 10, 10, 2, 1
        tolerance = 1e-6
        ocp, x, u = integrator_control_problem(T, u_max, x0, MultipleShooting(N=N,M=3,intg='rk'))
        sol = ocp.solve()
        ts, xs = sol.sample(x, grid='integrator')
        ts, us = sol.sample(u, grid='integrator')
        ts, uxs = sol.sample(u * x, grid='integrator')
        ts, tss = sol.sample(ocp.t, grid='integrator')

        t_exact = np.linspace(0, T, N * 3 + 1)
        x_exact = np.linspace(1, x0 - 10 * u_max, N * 3 + 1)
        u_exact = np.ones(N * 3 + 1) * (-u_max)

        assert_allclose(ts, t_exact, atol=tolerance)
        assert_allclose(xs, x_exact, atol=tolerance)
        assert_allclose(us, u_exact, atol=tolerance)
        assert_allclose(uxs, u_exact * x_exact, atol=tolerance)
        assert_allclose(ts, tss, atol=tolerance)

        tsa, tsb = sol.sample(ocp.t, grid='integrator')
        assert_allclose(tsa, tsb, atol=tolerance)

        s = ocp.sampler(ocp.t)
        assert_allclose(s(sol.gist,ts), tss, atol=tolerance)

    def test_intg_refine(self):
        for M in [1, 2]:
          for method in [DirectCollocation(N=2,M=M), MultipleShooting(N=2,M=M,intg='rk'), SingleShooting(N=2,M=M,intg='rk')] + [SplineMethod(N=2)] if M==1 else []:
            ocp, p, v, u = bang_bang_problem(method)
            sol = ocp.solve()
            tolerance = 1e-6
            grid = "integrator"
            if isinstance(method,SplineMethod):
              grid = "control"

            ts, ps = sol.sample(p, grid=grid, refine=10)

            ps_ref = np.hstack(((0.5*np.linspace(0,1, 10*M+1)**2)[:-1],np.linspace(0.5,1.5,10*M+1)-0.5*np.linspace(0,1, 10*M+1)**2)) 
            assert_allclose(ps, ps_ref, atol=tolerance)

            ts_ref = np.linspace(0, 2, 10*2*M+1)
            assert_allclose(ts, ts_ref, atol=tolerance)

            ts, vs = sol.sample(v, grid=grid, refine=10)
            assert_allclose(ts, ts_ref, atol=tolerance)

            vs_ref = np.hstack((np.linspace(0,1, 10*M+1)[:-1],np.linspace(1,0, 10*M+1))) 
            assert_allclose(vs, vs_ref, atol=tolerance)


            u_ref = np.array([1.0]*M*10+[-1.0]*(M*10+1))
            ts, us = sol.sample(u, grid=grid, refine=10)
            assert_allclose(us, u_ref, atol=tolerance)

            ts, tss = sol.sample(ocp.t, grid=grid, refine=10)
            assert_allclose(ts, tss, atol=tolerance)

    def test_shapes(self):
      N = 3
      M = 5
      R = 2
      for xshape in [(), (7,), (7, 11), (1, 1), (7, 1), (1, 7)]:
        ocp = Ocp(T=1)
        target_shape = tuple([e for e in xshape if e!=1])
        x = ocp.state(*xshape)
        ocp.set_der(x,DM.ones(*xshape))
        X0 = DM.rand(*xshape)
        ocp.subject_to(ocp.at_t0(x)==X0)
        ocp.solver('ipopt')

        X0_target = np.array(X0).reshape(target_shape)

        degree = 2
        ocp.method(DirectCollocation(N=N,M=M,degree=2))

        t, X = ocp.sample(x,grid='control')
        self.assertTrue(t.is_column())
        t, X = ocp.sample(x,grid='integrator')
        self.assertTrue(t.is_column())
        t, X = ocp.sample(x,grid='integrator',refine=R)
        self.assertTrue(t.is_column())
        t, X = ocp.sample(x,grid='integrator_roots')
        self.assertTrue(t.is_column())
        sol = ocp.solve()
        
        sampler_casadi = ocp.sampler('sampler',[x])
        sampler_numpy1 = ocp.sampler([x])
        sampler_numpy2 = ocp.sampler(x)

        for kwargs, tdim in [ (dict(grid='control'),N+1),
                              (dict(grid='integrator'),N*M+1),
                              (dict(grid='integrator',refine=R),N*M*R+1),
                              (dict(grid='integrator_roots'),N*M*degree)]:
          t, X = sol.sample(x,**kwargs)
          self.assertEqual(t.shape[0],tdim)
          self.assertEqual(X.shape[0],tdim)
          assert_allclose(X.shape[1:],target_shape)
          for i in range(tdim):
            assert_allclose(X[i,...],X0_target+t[i])
          if xshape != (1,7):
            assert_allclose(sampler_casadi(sol.gist, DM(t).T),hcat(list(X)))
          assert_allclose(sampler_numpy2(sol.gist, t[1]), X[1,...])
          assert_allclose(sampler_numpy1(sol.gist, t[1])[0], X[1,...])
          assert_allclose(sampler_numpy2(sol.gist, t), X)
          assert_allclose(sampler_numpy1(sol.gist, t)[0], X)


          sampler_casadi_sol = sol.sampler('sampler',[x])
          sampler_numpy1_sol = sol.sampler([x])
          sampler_numpy2_sol = sol.sampler(x)

          if xshape != (1,7):
            assert_allclose(sampler_casadi_sol(DM(t).T),hcat(list(X)))
          assert_allclose(sampler_numpy2_sol(t[1]), X[1,...])
          assert_allclose(sampler_numpy1_sol(t[1])[0], X[1,...])
          assert_allclose(sampler_numpy2_sol(t), X)
          assert_allclose(sampler_numpy1_sol(t)[0], X)

if __name__ == '__main__':
    unittest.main()



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import os
import sys

in_matlab = "libmwbuffer" in sys.modules.keys()

if in_matlab:
  try:
    dlopen_flags = sys.getdlopenflags()
    # Fixes crash of numpy in Matlab
    sys.setdlopenflags(10) # RTLD_DEEPBIND(8) | RTLD_NOW(2)
  except:
    pass

from .multiple_shooting import MultipleShooting
from .ocp import Ocp
from .stage import Stage
from .direct_method import DirectMethod
from .direct_collocation import DirectCollocation
from .spline_method import SplineMethod

from .single_shooting import SingleShooting
from .freetime import FreeTime
from .sampling_method import FreeGrid, UniformGrid, GeometricGrid, DensityGrid, DenseEdgesGrid
from .grouping_techniques import LseGroup
from .solution import OcpSolution
from .external.manager import external_method
from .casadi_helpers import rockit_pickle_context, rockit_unpickle_context, interface_simulink


try:
  matlab_path = os.path.realpath(os.path.dirname(os.path.realpath(__file__)))
except:
  matlab_path = "not_found"

if in_matlab:
  try:
    sys.setdlopenflags(dlopen_flags)
    del dlopen_flags
  except:
    pass

del in_matlab



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from typing import Union

import casadi as ca
import numpy as np

from casadi import MX, SX

from .utils import is_empty, casadi_length, is_column
from .acados_dims import AcadosOcpDims, AcadosSimDims


class AcadosModel():
    """
    Class containing all the information to code generate the external CasADi functions
    that are needed when creating an acados ocp solver or acados integrator.
    Thus, this class contains:

    a) the :py:attr:`name` of the model,
    b) all CasADi variables/expressions needed in the CasADi function generation process.
    """
    def __init__(self):
        ## common for OCP and Integrator
        self.name = None
        """The model name is used for code generation. Type: string. Default: :code:`None`"""
        self.x = []
        """CasADi variable describing the state of the system; Default: :code:`[]`"""
        self.xdot = []
        """CasADi variable describing the derivative of the state wrt time; Default: :code:`[]`"""
        self.u = []
        """CasADi variable describing the input of the system; Default: :code:`[]`"""
        self.z = []
        """CasADi variable describing the algebraic variables of the DAE; Default: :code:`[]`"""
        self.p = []
        """CasADi variable describing parameters of the DAE; Default: :code:`[]`"""
        self.t = []
        """
        CasADi variable representing time t in functions; Default: :code:`[]`
        NOTE:
        - For integrators, the start time has to be explicitly set via :py:attr:`acados_template.AcadosSimSolver.set`('t0').
        - For OCPs, the start time is set to 0. on each stage.
        The time dependency can be used within cost formulations and is relevant when cost integration is used.
        Start times of shooting intervals can be added using parameters.
        """
        self.p_global = []
        """
        CasADi variable containing global parameters.
        This feature can be used to precompute expensive terms which only depend on these parameters, e.g. spline coefficients, when p_global are underlying data points.
        Only supported for OCP solvers.
        Updating these parameters can be done using :py:attr:`acados_template.acados_ocp_solver.AcadosOcpSolver.set_p_global_and_precompute_dependencies(values)`.
        NOTE: this is only supported with CasADi beta release https://github.com/casadi/casadi/releases/tag/nightly-se
        Default: :code:`[]`
        """

        ## dynamics
        self.f_impl_expr = []
        r"""
        CasADi expression for the implicit dynamics :math:`f_\text{impl}(\dot{x}, x, u, z, p) = 0`.
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.integrator_type` == 'IRK'.
        Default: :code:`[]`
        """
        self.f_expl_expr = []
        r"""
        CasADi expression for the explicit dynamics :math:`\dot{x} = f_\text{expl}(x, u, p)`.
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.integrator_type` == 'ERK'.
        Default: :code:`[]`
        """
        self.disc_dyn_expr = []
        r"""
        CasADi expression for the discrete dynamics :math:`x_{+} = f_\text{disc}(x, u, p)`.
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.integrator_type` == 'DISCRETE'.
        Default: :code:`[]`
        """

        self.dyn_ext_fun_type = 'casadi'
        """type of external functions for dynamics module; 'casadi' or 'generic'; Default: 'casadi'"""
        self.dyn_generic_source = None
        """name of source file for discrete dynamics, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""
        self.dyn_disc_fun_jac_hess = None
        """name of function discrete dynamics + jacobian and hessian, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""
        self.dyn_disc_fun_jac = None
        """name of function discrete dynamics + jacobian, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""
        self.dyn_disc_fun = None
        """name of function discrete dynamics, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""

        self.dyn_impl_dae_fun_jac = None
        """name of source files for implicit DAE function value and jacobian, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""
        self.dyn_impl_dae_jac = None
        """name of source files for implicit DAE jacobian, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""
        self.dyn_impl_dae_fun = None
        """name of source files for implicit DAE function value, only relevant if :code:`dyn_ext_fun_type` is :code:`'generic'`; Default: :code:`None`"""

        # for GNSF models
        self.gnsf_nontrivial_f_LO = 1
        """GNSF: Flag indicating whether GNSF stucture has nontrivial f."""
        self.gnsf_purely_linear = 0
        """GNSF: Flag indicating whether GNSF stucture is purely linear."""


        ### for OCP only.
        # NOTE: These could be moved to cost / constraints

        # constraints at initial stage
        self.con_h_expr_0 = []
        """CasADi expression for the initial constraint :math:`h^0`; Default: :code:`[]`"""
        self.con_phi_expr_0 = []
        r"""CasADi expression for the terminal constraint :math:`\phi_0`; Default: :code:`[]`"""
        self.con_r_expr_0 = []
        r"""CasADi expression for the terminal constraint :math:`\phi_0(r)`,
        dummy input for outer function; Default: :code:`[]`"""
        self.con_r_in_phi_0 = []
        r"""CasADi expression for the terminal constraint :math:`\phi_0(r)`, input for outer function; Default: :code:`[]`"""


        # path constraints
        # BGH(default): lh <= h(x, u) <= uh
        self.con_h_expr = []
        """CasADi expression for the constraint :math:`h`; Default: :code:`[]`"""
        # BGP(convex over nonlinear): lphi <= phi(r(x, u)) <= uphi
        self.con_phi_expr = []
        """CasADi expression for the constraint phi; Default: :code:`[]`"""
        self.con_r_expr = []
        """CasADi expression for the constraint phi(r),
        dummy input for outer function; Default: :code:`[]`"""
        self.con_r_in_phi = []
        r"""CasADi expression for the terminal constraint :math:`\phi(r)`,
        input for outer function; Default: :code:`[]`"""

        # terminal
        self.con_h_expr_e = []
        """CasADi expression for the terminal constraint :math:`h^e`; Default: :code:`[]`"""
        self.con_phi_expr_e = []
        r"""CasADi expression for the terminal constraint :math:`\phi_e`; Default: :code:`[]`"""
        self.con_r_expr_e = []
        r"""CasADi expression for the terminal constraint :math:`\phi_e(r)`,
        dummy input for outer function; Default: :code:`[]`"""
        self.con_r_in_phi_e = []
        r"""CasADi expression for the terminal constraint :math:`\phi_e(r)`, input for outer function; Default: :code:`[]`"""

        # cost
        self.cost_y_expr = []
        """CasADi expression for nonlinear least squares; Default: :code:`[]`"""
        self.cost_y_expr_e = []
        """CasADi expression for nonlinear least squares, terminal; Default: :code:`[]`"""
        self.cost_y_expr_0 = []
        """CasADi expression for nonlinear least squares, initial; Default: :code:`[]`"""
        self.cost_expr_ext_cost = []
        """CasADi expression for external cost; Default: :code:`[]`"""
        self.cost_expr_ext_cost_e = []
        """CasADi expression for external cost, terminal; Default: :code:`[]`"""
        self.cost_expr_ext_cost_0 = []
        """CasADi expression for external cost, initial; Default: :code:`[]`"""
        self.cost_expr_ext_cost_custom_hess = []
        """CasADi expression for custom hessian (only for external cost); Default: :code:`[]`"""
        self.cost_expr_ext_cost_custom_hess_e = []
        """CasADi expression for custom hessian (only for external cost), terminal; Default: :code:`[]`"""
        self.cost_expr_ext_cost_custom_hess_0 = []
        """CasADi expression for custom hessian (only for external cost), initial; Default: :code:`[]`"""

        ## CONVEX_OVER_NONLINEAR convex-over-nonlinear cost: psi(y(x, u, p) - y_ref; p)
        self.cost_psi_expr_0 = []
        r"""
        CasADi expression for the outer loss function :math:`\psi(r - yref, t, p)`, initial; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type_0` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_psi_expr = []
        r"""
        CasADi expression for the outer loss function :math:`\psi(r - yref, t, p)`; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_psi_expr_e = []
        r"""
        CasADi expression for the outer loss function :math:`\psi(r - yref, p)`, terminal; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type_e` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_r_in_psi_expr_0 = []
        r"""
        CasADi symbolic input variable for the argument :math:`r` to the outer loss function :math:`\psi(r, t, p)`, initial; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type_0` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_r_in_psi_expr = []
        r"""
        CasADi symbolic input variable for the argument :math:`r` to the outer loss function :math:`\psi(r, t, p)`; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_r_in_psi_expr_e = []
        r"""
        CasADi symbolic input variable for the argument :math:`r` to the outer loss function :math:`\psi(r, p)`, terminal; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type_e` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_conl_custom_outer_hess_0 = []
        """
        CasADi expression for the custom hessian of the outer loss function (only for convex-over-nonlinear cost), initial; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type_0` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_conl_custom_outer_hess = []
        """
        CasADi expression for the custom hessian of the outer loss function (only for convex-over-nonlinear cost); Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.cost_conl_custom_outer_hess_e = []
        """
        CasADi expression for the custom hessian of the outer loss function (only for convex-over-nonlinear cost), terminal; Default: :code:`[]`
        Used if :py:attr:`acados_template.acados_ocp_options.AcadosOcpOptions.cost_type_e` is 'CONVEX_OVER_NONLINEAR'.
        """
        self.nu_original = None # TODO: remove? only used by benchmark
        """
        Number of original control inputs (before polynomial control augmentation); Default: :code:`None`
        """
        self.t0 = None # TODO: remove? only used by benchmark
        """CasADi variable representing the start time of an interval; Default: :code:`None`"""
        self.__x_labels = None
        self.__u_labels = None
        self.__t_label = "t"

    @property
    def x_labels(self):
        """Contains list of labels for the states. Default: :code:`None`"""
        if self.__x_labels is None:
            return [f"x{i}" for i in range(self.x.size()[0])]
        else:
            return self.__x_labels

    @x_labels.setter
    def x_labels(self, x_labels):
        self.__x_labels = x_labels


    @property
    def u_labels(self):
        """Contains list of labels for the controls. Default: :code:`None`"""
        if self.__u_labels is None:
            return [f"x{i}" for i in range(self.x.size()[0])]
        else:
            return self.__u_labels

    @u_labels.setter
    def u_labels(self, u_labels):
        self.__u_labels = u_labels

    @property
    def t_label(self):
        """Label for the time variable. Default: :code:'t'"""
        return self.__t_label

    @t_label.setter
    def t_label(self, t_label):
        self.__t_label = t_label


    def get_casadi_symbol(self):
        if isinstance(self.x, MX):
            return MX.sym
        elif isinstance(self.x, SX):
            return SX.sym
        else:
            raise Exception(f"model.x must be casadi.SX or casadi.MX, got {type(self.x)}")

    def get_casadi_zeros(self):
        if isinstance(self.x, MX):
            return MX.zeros
        elif isinstance(self.x, SX):
            return SX.zeros
        else:
            raise Exception(f"model.x must be casadi.SX or casadi.MX, got {type(self.x)}")


    def make_consistent(self, dims: Union[AcadosOcpDims, AcadosSimDims]) -> None:

        casadi_symbol = self.get_casadi_symbol()

        # nx
        if is_empty(self.x):
            raise Exception("model.x must be defined")
        else:
            dims.nx = casadi_length(self.x)

        if is_empty(self.xdot):
            self.xdot = casadi_symbol('xdot', dims.nx, 1)
        else:
            if casadi_length(self.xdot) != dims.nx:
                raise Exception(f"model.xdot must have length nx = {dims.nx}, got {casadi_length(self.xdot)}")

        # nu
        if is_empty(self.u):
            dims.nu = 0
            self.u = casadi_symbol('u', 0, 1)
        else:
            dims.nu = casadi_length(self.u)

        # nz
        if is_empty(self.z):
            dims.nz = 0
            self.z = casadi_symbol('z', 0, 1)
        else:
            dims.nz = casadi_length(self.z)

        # np
        if is_empty(self.p):
            dims.np = 0
            self.p = casadi_symbol('p', 0, 1)
        else:
            dims.np = casadi_length(self.p)

        # np_global
        if is_empty(self.p_global):
            dims.np_global = 0
            self.p_global = casadi_symbol('p_global', 0, 1)
        else:
            dims.np_global = casadi_length(self.p_global)

        # sanity checks
        for symbol, name in [(self.x, 'x'), (self.xdot, 'xdot'), (self.u, 'u'), (self.z, 'z'), (self.p, 'p'), (self.p_global, 'p_global')]:
            if not isinstance(symbol, (ca.MX, ca.SX)):
                raise Exception(f"model.{name} must be casadi.MX, casadi.SX got {type(symbol)}")
            if not symbol.is_valid_input():
                raise Exception(f"model.{name} must be valid CasADi symbol, got {symbol}")

        # p_global
        if not is_empty(self.p_global):
            if isinstance(dims, AcadosSimDims):
                raise Exception("model.p_global is only supported for OCPs")
            if any(ca.which_depends(self.p_global, self.p)):
                raise Exception(f"model.p_global must not depend on model.p, got p_global ={self.p_global}, p = {self.p}")

        # model output dimension nx_next: dimension of the next state
        if isinstance(dims, AcadosOcpDims):
            if not is_empty(self.disc_dyn_expr):
                dims.nx_next = casadi_length(self.disc_dyn_expr)
            else:
                dims.nx_next = casadi_length(self.x)

        if not is_empty(self.f_impl_expr):
            if casadi_length(self.f_impl_expr) != (dims.nx + dims.nz):
                raise Exception(f"model.f_impl_expr must have length nx + nz = {dims.nx} + {dims.nz}, got {casadi_length(self.f_impl_expr)}")
        if not is_empty(self.f_expl_expr):
            if casadi_length(self.f_expl_expr) != dims.nx:
                raise Exception(f"model.f_expl_expr must have length nx = {dims.nx}, got {casadi_length(self.f_expl_expr)}")

        return


    def substitute(self, var: Union[ca.SX, ca.MX], expr_new: Union[ca.SX, ca.MX]) -> None:
        """
        Substitutes the variables var with expr_new in all symbolic CasADi expressions within AcadosModel
        """
        for attr, value in self.__dict__.items():
            if isinstance(value, (ca.SX, ca.MX)):
                new = ca.substitute(value, var, expr_new)
                setattr(self, attr, new)
        return


    def augment_model_with_polynomial_control(self, degree: int) -> None:
        print("Deprecation warning: augment_model_with_polynomial_control() is deprecated and has been renamed to reformulate_with_polynomial_control().")
        self.reformulate_with_polynomial_control(degree=degree)


    def reformulate_with_polynomial_control(self, degree: int) -> None:
        r"""
        Augment the model with polynomial control.

        Replace the original control input :math:`u` with a polynomial control input :math:`v_{\text{poly}} = \sum_{i=0}^d u_i t^i`
        New controls are :math:`u_0, \dots, u_d`.

        NOTE: bounds on controls are not changed in this function.

        :param degree: degree of the polynomial control
        :type degree: int
        """
        if self.u is None:
            raise Exception('model.u must be defined')
        if self.nu_original is not None:
            raise Exception('model.u has already been augmented')

        casadi_symbol = self.get_casadi_symbol()

        # add time to model
        if self.t == []:
            self.t = casadi_symbol('t')

        t = self.t

        u_old = self.u
        nu_original = casadi_length(self.u)

        u_coeff = casadi_symbol('u_coeff', (degree+1) * nu_original)
        u_new = np.zeros((nu_original, 1))
        for i in range(degree+1):
            u_new += t ** i * u_coeff[i*nu_original:(i+1)*nu_original]

        evaluate_polynomial_u_fun = ca.Function("evaluate_polynomial_u", [u_coeff, t], [u_new])

        self.substitute(u_old, u_new)

        self.u = u_coeff
        self.nu_original = nu_original
        self.p = ca.vertcat(self.p)

        # update name
        self.name = self.name + f"_d{degree}"

        return evaluate_polynomial_u_fun



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import os
import importlib

def external_method(name, **kwargs):
    try:
        external = importlib.import_module('.external.' + name, package='rockit')
    except ModuleNotFoundError:
        externals = []
        with os.scandir(os.path.dirname(os.path.realpath(__file__))) as it:
            for entry in it:
                if not entry.name.startswith('.') and not entry.name.startswith('_') and not entry.is_file():
                    externals.append(entry.name)
        externals = ",".join(["'%s'" % e for e in externals])
        raise Exception("external '%s' not found. Available: %s." % (name, externals))
    return external.method(**kwargs)

    


# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from .acados_sim_solver import AcadosSimSolver
from .acados_sim import AcadosSim
from typing import List
from ctypes import (POINTER, c_int, c_void_p)


class AcadosSimBatchSolver():
    """
    Batch Integrator for parallel integration.

        :param sim: type :py:class:`~acados_template.acados_sim.AcadosSim`
        :param N_batch: batch size, positive integer
        :param json_file: Default: 'acados_sim.json'
        :verbose: bool, default: True
    """

    __sim_solvers : List[AcadosSimSolver]

    def __init__(self, sim: AcadosSim, N_batch: int, json_file: str = 'acados_sim.json', verbose: bool=True):

        if not isinstance(N_batch, int) or N_batch <= 0:
            raise Exception("AcadosSimBatchSolver: argument N_batch should be a positive integer.")

        self.__N_batch = N_batch
        self.__sim_solvers = [AcadosSimSolver(sim, json_file=json_file, build=n==0, generate=n==0, verbose=verbose) for n in range(self.N_batch)]

        self.__shared_lib = self.sim_solvers[0].shared_lib
        self.__model_name = self.sim_solvers[0].model_name
        self.__sim_solvers_pointer = (c_void_p * self.N_batch)()

        for i in range(self.N_batch):
            self.__sim_solvers_pointer[i] = self.sim_solvers[i].capsule

        getattr(self.__shared_lib, f"{self.__model_name}_acados_sim_batch_solve").argtypes = [POINTER(c_void_p), c_int]
        getattr(self.__shared_lib, f"{self.__model_name}_acados_sim_batch_solve").restype = c_void_p

        if not self.sim_solvers[0].acados_lib_uses_omp:
            print("Warning: Please compile the acados shared library with openmp and the number of threads set to 1, i.e. with the flags -DACADOS_WITH_OPENMP=ON -DACADOS_NUM_THREADS=1.")


    def solve(self):
        """
        Solve the simulation problem with current input for all `N_batch` integrators.
        """
        getattr(self.__shared_lib, f"{self.__model_name}_acados_sim_batch_solve")(self.__sim_solvers_pointer, self.__N_batch)


    @property
    def sim_solvers(self):
        """List of AcadosSimSolvers."""
        return self.__sim_solvers

    @property
    def N_batch(self):
        """Batch size."""
        return self.__N_batch




#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

import casadi as cs
import os
from casadi import *
from collections import defaultdict, OrderedDict
from contextlib import contextmanager

def get_ranges_dict(list_expr):
    ret = HashDict()
    offset = 0
    for e in list_expr:
        next_offset = offset+e.nnz()
        ret[e] = list(range(offset, next_offset))
        offset = next_offset
    return ret

def reinterpret_expr(expr, symbols_from, symbols_to):
    """
    .. code-block:: python
    
        x = MX.sym("x")
        y = MX.sym("y")

        z = -(x*y**2+2*x)**4<0

        print(reinterpret_expr(z,[y],[sin(y)]))
    """

    try:
        f = Function('f', symbols_from, [expr], {"allow_free":True})
    except:
        f = Function('f', symbols_from, [expr])

    # Work vector
    work = [None for i in range(f.sz_w())]

    output_val = [None]

    # Loop over the algorithm
    for k in range(f.n_instructions()):

        # Get the atomic operation
        op = f.instruction_id(k)

        o = f.instruction_output(k)
        i = f.instruction_input(k)

        if(op == OP_CONST):
            v = f.instruction_MX(k).to_DM()
            work[o[0]] = v
        else:
            if op == OP_INPUT:
                work[o[0]] = symbols_to[i[0]]
            elif op == OP_OUTPUT:
                output_val[o[0]] = work[i[0]]
            elif op == OP_ADD:
                work[o[0]] = work[i[0]] + work[i[1]]
            elif op == OP_TWICE:
                work[o[0]] = 2 * work[i[0]]
            elif op == OP_SUB:
                work[o[0]] = work[i[0]] - work[i[1]]
            elif op == OP_MUL:
                work[o[0]] = work[i[0]] * work[i[1]]
            elif op == OP_MTIMES:
                work[o[0]] = np.dot(work[i[1]], work[i[2]]) + work[i[0]]
            elif op == OP_PARAMETER:
                work[o[0]] = f.instruction_MX(k)
            elif op == OP_SQ:
                work[o[0]] = work[i[0]]**2
            elif op == OP_LE:
                work[o[0]] = work[i[0]] <= work[i[1]]
            elif op == OP_LT:
                work[o[0]] = work[i[0]] < work[i[1]]
            elif op == OP_NEG:
                work[o[0]] = -work[i[0]]
            elif op == OP_CONSTPOW:
                work[o[0]] = work[i[0]]**work[i[1]]
            elif op == OP_VERTCAT:
                work[o[0]] = vcat([work[e] for e in i])
            elif op == OP_CALL:
                cn = f.instruction_MX(k)
                fn = cn.which_function()
                res = fn(*[work[e] for e in i])
                for i,e in enumerate(o):
                    work[e] = res[i]
            else:
                try:
                    e = f.instruction_MX(k)
                    res = e.eval_mx([work[e] for e in i])
                    for i,e in enumerate(o):
                        work[e] = res[i]
                except:
                    print('Unknown operation: ', op)

                    print('------')
                    print('Evaluated ' + str(f))
                    raise Exception()

    return output_val[0]


def get_meta(base=None):
    if base is not None: return base
    # Construct meta-data
    import sys
    import os
    try:
        frame = sys._getframe(2)
        meta = {"stacktrace": [{"file":os.path.abspath(frame.f_code.co_filename),"line":frame.f_lineno,"name":frame.f_code.co_name} ] }
    except:
        meta = {"stacktrace": []}
    return meta

def merge_meta(a, b):
    if b is None:
        return a
    if a is None:
        return b
    from copy import deepcopy
    res = deepcopy(a)
    res["stacktrace"] += b["stacktrace"] 
    return res

def single_stacktrace(m):
    from copy import deepcopy
    m = deepcopy(m)
    m["stacktrace"] = m["stacktrace"][0]
    return m

def reshape_number(target, value):
    if not isinstance(value,cs.MX):
        value = cs.DM(value)
    if value.is_scalar():
        value = cs.DM.ones(target.shape)*value
    return value

def is_numeric(expr):
    if isinstance(expr,cs.DM):
        return True
    elif isinstance(expr,np.ndarray):
        return True
    try:
        expr = evalf(expr)
        return True
    except:
        return False

def DM2numpy(dm, expr_shape, tdim=None):
    if tdim is None:
        return np.array(dm).squeeze()
    expr_prod = expr_shape[0]*expr_shape[1]

    target_shape = (tdim,)+tuple([e for e in expr_shape if e!=1])

    res = np.array(dm).reshape(expr_shape[0], tdim, expr_shape[1])
    res = np.transpose(res,[1,0,2])
    res = res.reshape(target_shape)
    return res

class HashWrap:
    def __init__(self, arg):
        assert not isinstance(arg,HashWrap)
        self.arg = arg

    def __hash__(self):
        return hash(self.arg)

    def __eq__(self, b):
        # key in dict
        b_arg = b
        if isinstance(b,HashWrap): b_arg = b.arg
        if self.arg is b_arg: return True
        r = self.arg==b_arg
        return r.is_one()

    def __str__(self):
        return self.arg.__str__()
    def __repr__(self):
        return self.arg.__repr__()

class HashDict(dict):
    def __init__(self,*args,**kwargs):
        r = dict(*args,**kwargs)
        dict.__init__(self)
        for k,v in r.items():
            self[k] = v
    def __getitem__(self, k):
        return dict.__getitem__(self, HashWrap(k))
    def __setitem__(self, k, v):
        return dict.__setitem__(self, HashWrap(k), v)
    def keys(self):
        for k in dict.keys(self):
            yield k.arg
    def items(self):
        for k,v in dict.items(self):
            yield k.arg, v
    __iter__ = keys
    def __copy__(self):
        r = HashDict()
        for k,v in self.items():
            r[k] = v
        return r

class HashList(list):
    def __init__(self,*args,**kwargs):
        r = list(*args,**kwargs)
        list.__init__(self)
        for v in r:
            self.append(v)
        self._stored = set()
    def append(self, item):
        list.append(self, item)
        self._stored.add(HashWrap(item))
    def __contains__(self, item):
        return item in self._stored
    def __copy__(self):
        r = HashList()
        for v in self:
            r.append(v)
        return r

class HashDefaultDict(defaultdict):
    def __init__(self,default_factory=None, *args,**kwargs):
        r = defaultdict(default_factory,*args,**kwargs)
        defaultdict.__init__(self)
        for k,v in r.items():
            self[k] = v
    def __getitem__(self, k):
        return defaultdict.__getitem__(self, HashWrap(k))
    def __setitem__(self, k, v):
        return defaultdict.__setitem__(self, HashWrap(k), v)
    def keys(self):
        ret = []
        for k in defaultdict.keys(self):
            ret.append(k.arg)
        return ret
    def items(self):
        for k,v in defaultdict.items(self):
            yield k.arg, v
    def __iter__(self):
        for k in defaultdict.__iter__(self):
            yield k.arg
    def __copy__(self):
        r = HashDefaultDict(self.default_factory)
        for k,v in self.items():
            r[k] = v
        return r

class HashOrderedDict(OrderedDict):
    def __init__(self,*args,**kwargs):
        r = OrderedDict(*args,**kwargs)
        OrderedDict.__init__(self)
        for k,v in r.items():
            self[k] = v
    def __getitem__(self, k):
        return OrderedDict.__getitem__(self, HashWrap(k))
    def __setitem__(self, k, v):
        return OrderedDict.__setitem__(self, HashWrap(k), v)
    def keys(self):
        ret = []
        for k in self:
            ret.append(k)
        return ret
    def items(self):
        for k in self:
            yield k, self[k]
    def __iter__(self):
        for k in OrderedDict.__iter__(self):
            yield k.arg
    def __copy__(self):
        r = HashOrderedDict()
        for k,v in self.items():
            r[k] = v
        return r


def for_all_primitives(expr, rhs, callback, msg, rhs_type=MX):
    if expr.is_symbolic():
        callback(expr, rhs)
        return
    if expr.is_valid_input():
        # Promote to correct size
        rhs = rhs_type(rhs)
        if rhs.is_scalar() and not expr.is_scalar():
            rhs = rhs*rhs_type(expr.sparsity())
        rhs = rhs[expr.sparsity()]
        rhs = vec(rhs)
        prim = expr.primitives()
        rhs = rhs.nz
        offset = 0
        for p in prim:
            callback(p, rhs_type(p.sparsity(), rhs[offset:offset+p.nnz()]))
            offset += p.nnz()
    else:
        raise Exception(msg)
        

class Node:
  def __init__(self,val):
    self.val = val
    self.nodes = []

class AutoBrancher:
  OPEN = 0
  DONE = 1
  def __init__(self):
    self.root = Node(AutoBrancher.OPEN)
    self.trace = [self.root]

  @property
  def current(self):
    return self.trace[-1]

  def branch(self, alternatives = [True, False]):
    alternatives = list(alternatives)
    nodes = self.current.nodes
    if len(nodes)==0:
      nodes += [None]*len(alternatives)
    for i,n in enumerate(nodes):
      if n is None:
        nodes[i] = Node(AutoBrancher.OPEN)
        self.trace.append(nodes[i])
        self.this_branch.append(alternatives[i])
        return alternatives[i]
      else:
        if n.val == AutoBrancher.OPEN:
          self.trace.append(nodes[i])
          self.this_branch.append(alternatives[i])
          return alternatives[i]

  def __iter__(self):
    cnt = 0
    
    while self.root.val==AutoBrancher.OPEN:
      self.this_branch = []
      cnt+=1
      yield self
      # Indicate that current leaf is done
      self.current.val = AutoBrancher.DONE
      # Close leaves when subleaves are done
      for n in reversed(self.trace[:-1]):
        finished = True
        for e in n.nodes:
          finished = finished and e and e.val==AutoBrancher.DONE
        if finished:
          n.val = AutoBrancher.DONE
      # Reset trace
      self.trace = [self.root]
      print("Evaluated branch",self.this_branch)
    print("Evaluated",cnt,"branches")

def vvcat(arg):
    if len(arg)==0:
        return cs.MX(0,1)
    else:
        return cs.vvcat(arg)

def vcat(arg):
    if len(arg)==0:
        return cs.MX(0,1)
    else:
        return cs.vcat(arg)

def prepare_build_dir(build_dir_abs):
    import os
    import shutil
    os.makedirs(build_dir_abs,exist_ok=True)
    # Clean directory (but do not delete it,
    # since this confuses open shells in Linux (e.g. bash, Matlab)
    for filename in os.listdir(build_dir_abs):
      file_path = os.path.join(build_dir_abs, filename)
      try:
        if os.path.isfile(file_path) or os.path.islink(file_path):
          os.unlink(file_path)
        elif os.path.isdir(file_path):
          shutil.rmtree(file_path)
      except:
        pass
    


class ConstraintInspector:
    def __init__(self, method, stage):
        self.opti = Opti()

        self.X = self.opti.variable(*stage.x.shape)
        self.U = self.opti.variable(*stage.u.shape)
        self.V = self.opti.variable(*stage.v.shape)
        self.P = self.opti.parameter(*stage.p.shape)
        self.t = self.opti.variable()
        self.T = self.opti.variable()

        offsets = list(stage._offsets.keys())

        self.offsets = []
        for e in offsets:
            self.offsets.append(self.opti.variable(*e.shape))

        self.raw = [stage.x,stage.u,stage.p,stage.t, method.v]+offsets
        self.optivar = [self.X, self.U, self.P, self.t, self.V]+self.offsets

        if method.free_time:
            self.raw += [stage.T]
            self.optivar += [self.T]

        self.method = method
    
    def finalize(self):
        if hasattr(self.method,'signals'):
            # Need to a add signal derivative symbols
            extra_vars = []
            extra_pars = []
            for k,v in self.method.signals.items():
                if v.derivative_of is None: continue
                if v.parametric:
                    extra_pars.append(v.symbol)
                else:
                    extra_vars.append(v.symbol)
            raw_var = vvcat(extra_vars)
            raw_par = vvcat(extra_pars)
            self.raw += [raw_var,raw_par]
            self.optivar += [self.opti.variable(*raw_var.shape),self.opti.parameter(*raw_par.shape)]
   
        self.opti_advanced = self.opti.advanced

    def canon(self,expr):
        c = substitute([expr],self.raw,self.optivar)[0]
        mc = self.opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
        return substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw), mc


def linear_coeffs(expr, *args):
    """ Multi-argument extesion to CasADi linear_coeff"""
    J,c = linear_coeff(expr, vcat(args))
    cs = np.cumsum([0]+[e.numel() for e in args])
    return tuple([J[:,cs[i]:cs[i+1]] for i in range(len(args))])+(c,)

ca_classes = [cs.SX,cs.MX,cs.Function,cs.Sparsity,cs.DM,cs.Opti]

@contextmanager
def rockit_pickle_context():
    string_serializer = cs.StringSerializer()
    string_serializer.pack("preamble")
    string_serializer.encode()
    def __getstate__(self):
        if isinstance(self,cs.Opti):
            raise Exception("Opti cannot be serialized yet. Consider removing the transcribed problem.")
        string_serializer.pack(self)
        enc = string_serializer.encode()
        return {"s":enc}

    for c in ca_classes:
        setattr(c, '__getstate__', __getstate__)
        
    yield

    for c in ca_classes:
        delattr(c, '__getstate__')

@contextmanager
def rockit_unpickle_context():
    string_serializer = cs.StringSerializer()
    string_serializer.pack("preamble")
    string_deserializer = cs.StringDeserializer(string_serializer.encode())
    string_deserializer.unpack()
    def __setstate__(self,state):
        string_deserializer.decode(state["s"])
        s = string_deserializer.unpack()
        self.this = s.this

    for c in ca_classes:
        setattr(c, '__setstate__', __setstate__)
        
    yield

    for c in ca_classes:
        delattr(c, '__setstate__')
        
def interface_simulink(mdl,path='',block='',exclude_outputs='',exclude_inputs='',options=None,jacobian_options=None):
    opts = {}
    external_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'external')
    opts["config_args"] = [external_dir,'DiffSimulinkModel',\
        mdl,\
        'path',path, \
        'block',block, \
        'exclude_outputs',exclude_outputs, \
        'exclude_inputs',exclude_inputs,
		'cwd',os.getcwd()]
    opts["enable_fd"] = True

    import copy
    opts_nominal = copy.deepcopy(opts)
    opts_nominal["config_args"] += ["with_diff","false"]

    if os.name == 'nt':
        suffix = ".dll"
    else:
        suffix = ".so"
        
    libmatlab_ipc_base = os.path.join(cs.GlobalOptions.getCasadiPath(),'libmatlab_ipc')

    libmatlab_ipc_nominal = './libmatlab_ipc_nominal'+suffix
    libmatlab_ipc_jacobian = './libmatlab_ipc_jacobian'+suffix

    import shutil
    shutil.copy(libmatlab_ipc_base+suffix,libmatlab_ipc_nominal)
    shutil.copy(libmatlab_ipc_base+suffix,libmatlab_ipc_jacobian)

    F = cs.external('F',libmatlab_ipc_nominal,opts_nominal)
    J = cs.external('F',libmatlab_ipc_jacobian,opts)

    nx = F.size1_in('x')
    nu = F.size1_in('u')
    np = F.size1_in('p')
    ny = F.size1_out('y')

    x = cs.MX.sym('x',nx)
    u = cs.MX.sym('u',nu)
    p = cs.MX.sym('p',np)

    res = J(x=x,u=u,p=p)

    out_names = ['jac_dx_x','jac_dx_u','jac_y_x','jac_y_u']


    if jacobian_options is None:
        jacobian_options = {}
    jacobian_options.update({"never_inline":True,"enable_fd":False,"enable_forward":False,"enable_forward":False,"enable_jacobian":False})

    jac_F = cs.Function('jac_F', \
        [x,u,p,cs.MX(nx,1),cs.MX(ny,1)], \
        [res['jac_dx_x'],res['jac_dx_u'],cs.DM(nx,np),res['jac_y_x'],res['jac_y_u'],cs.DM(ny,np)], \
        ['x','u','p','out_dx','out_y'], \
        ['jac_dx_x','jac_dx_u','jac_dx_p','jac_y_x','jac_y_u','jac_y_p'], \
        jacobian_options)

    res = F(x=x,u=u,p=p)
    out_names = ['dx','y']

    if options is None:
        options = {}
    options.update({"never_inline":True,"enable_fd":False,"enable_forward":False,"enable_forward":False,"enable_jacobian":True,"jac_penalty":0,"custom_jacobian":jac_F})

    F = cs.Function('F', \
        [x,u,p], \
        [res[k] for k in out_names], \
        ['x','u','p'], \
        out_names, \
        options)

    return F




#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#
#   Author: Jonathan Frey: jonathanpaulfrey(at)gmail.com

from casadi import Function, jacobian, SX, horzcat, depends_on

from .determine_trivial_gnsf_transcription import determine_trivial_gnsf_transcription
from .detect_affine_terms_reduce_nonlinearity import (
    detect_affine_terms_reduce_nonlinearity,
)
from .reformulate_with_LOS import reformulate_with_LOS
from .reformulate_with_invertible_E_mat import reformulate_with_invertible_E_mat
from .structure_detection_print_summary import structure_detection_print_summary
from .check_reformulation import check_reformulation
from ..utils import is_empty


def detect_gnsf_structure(acados_ocp, transcribe_opts=None):

    ## Description
    # This function takes a CasADi implicit ODE or index-1 DAE model "model"
    # consisting of a CasADi expression f_impl in the symbolic CasADi
    # variables x, xdot, u, z, (and possibly parameters p), which are also part
    # of the model, as well as a model name.
    # It will create a struct "gnsf" containing all information needed to use
    # it with the gnsf integrator in acados.
    # Additionally it will create the struct "reordered_model" which contains
    # the permuted state vector and permuted f_impl, in which additionally some
    # functions, which were made part of the linear output system of the gnsf,
    # have changed signs.

    # Options: transcribe_opts is a Matlab struct consisting of booleans:
    #   print_info: if extensive information on how the model is processed
    #       is printed to the console.
    #   generate_gnsf_model: if the neccessary C functions to simulate the gnsf
    #       model with the acados implementation of the GNSF exploiting
    #       integrator should be generated.
    #   generate_gnsf_model: if the neccessary C functions to simulate the
    #       reordered model with the acados implementation of the IRK
    #       integrator should be generated.
    #   check_E_invertibility: if the transcription method should check if the
    #       assumption that the main blocks of the matrix gnsf.E are invertible
    #       holds. If not, the method will try to reformulate the gnsf model
    #       with a different model, such that the assumption holds.

    # acados_root_dir = getenv('ACADOS_INSTALL_DIR')

    if not is_empty(acados_ocp.model.p_global) and depends_on(acados_ocp.model.f_impl_expr, acados_ocp.model.p_global):
        Exception("GNSF does not support global parameters")

    ## load transcribe_opts
    if transcribe_opts is None:
        print("WARNING: GNSF structure detection called without transcribe_opts")
        print(" using default settings")
        print("")
        transcribe_opts = dict()

    if "print_info" in transcribe_opts:
        print_info = transcribe_opts["print_info"]
    else:
        print_info = 1
        print("print_info option was not set - default is true")

    if "detect_LOS" in transcribe_opts:
        detect_LOS = transcribe_opts["detect_LOS"]
    else:
        detect_LOS = 1
        if print_info:
            print("detect_LOS option was not set - default is true")

    if "check_E_invertibility" in transcribe_opts:
        check_E_invertibility = transcribe_opts["check_E_invertibility"]
    else:
        check_E_invertibility = 1
        if print_info:
            print("check_E_invertibility option was not set - default is true")

    ## Reformulate implicit index-1 DAE into GNSF form
    # (Generalized nonlinear static feedback)
    gnsf = determine_trivial_gnsf_transcription(acados_ocp, print_info)
    gnsf = detect_affine_terms_reduce_nonlinearity(gnsf, acados_ocp, print_info)

    if detect_LOS:
        gnsf = reformulate_with_LOS(acados_ocp, gnsf, print_info)

    if check_E_invertibility:
        gnsf = reformulate_with_invertible_E_mat(gnsf, acados_ocp, print_info)

    # detect purely linear model
    if gnsf["nx1"] == 0 and gnsf["nz1"] == 0 and gnsf["nontrivial_f_LO"] == 0:
        gnsf["purely_linear"] = 1
    else:
        gnsf["purely_linear"] = 0

    structure_detection_print_summary(gnsf, acados_ocp)
    check_reformulation(acados_ocp.model, gnsf, print_info)

    ## copy relevant fields from gnsf to model
    acados_ocp.model.get_matrices_fun = Function()
    dummy = acados_ocp.model.x[0]
    model_name = acados_ocp.model.name

    get_matrices_fun = Function(
        f"{model_name}_gnsf_get_matrices_fun",
        [dummy],
        [
            gnsf["A"],
            gnsf["B"],
            gnsf["C"],
            gnsf["E"],
            gnsf["L_x"],
            gnsf["L_xdot"],
            gnsf["L_z"],
            gnsf["L_u"],
            gnsf["A_LO"],
            gnsf["c"],
            gnsf["E_LO"],
            gnsf["B_LO"],
            gnsf["nontrivial_f_LO"],
            gnsf["purely_linear"],
            gnsf["ipiv_x"] + 1,
            gnsf["ipiv_z"] + 1,
            gnsf["c_LO"],
        ],
    )

    phi = gnsf["phi_expr"]
    y = gnsf["y"]
    uhat = gnsf["uhat"]
    p = gnsf["p"]

    jac_phi_y = jacobian(phi, y)
    jac_phi_uhat = jacobian(phi, uhat)

    phi_fun = Function(f"{model_name}_gnsf_phi_fun", [y, uhat, p], [phi])
    acados_ocp.model.phi_fun = phi_fun
    acados_ocp.model.phi_fun_jac_y = Function(
        f"{model_name}_gnsf_phi_fun_jac_y", [y, uhat, p], [phi, jac_phi_y]
    )
    acados_ocp.model.phi_jac_y_uhat = Function(
        f"{model_name}_gnsf_phi_jac_y_uhat", [y, uhat, p], [jac_phi_y, jac_phi_uhat]
    )

    x1 = acados_ocp.model.x[gnsf["idx_perm_x"][: gnsf["nx1"]]]
    x1dot = acados_ocp.model.xdot[gnsf["idx_perm_x"][: gnsf["nx1"]]]
    if gnsf["nz1"] > 0:
        z1 = acados_ocp.model.z[gnsf["idx_perm_z"][: gnsf["nz1"]]]
    else:
        z1 = SX.sym("z1", 0, 0)
    f_lo = gnsf["f_lo_expr"]
    u = acados_ocp.model.u
    acados_ocp.model.f_lo_fun_jac_x1k1uz = Function(
        f"{model_name}_gnsf_f_lo_fun_jac_x1k1uz",
        [x1, x1dot, z1, u, p],
        [
            f_lo,
            horzcat(
                jacobian(f_lo, x1),
                jacobian(f_lo, x1dot),
                jacobian(f_lo, u),
                jacobian(f_lo, z1),
            ),
        ],
    )

    acados_ocp.model.get_matrices_fun = get_matrices_fun

    size_gnsf_A = gnsf["A"].shape
    acados_ocp.dims.gnsf_nx1 = size_gnsf_A[1]
    acados_ocp.dims.gnsf_nz1 = size_gnsf_A[0] - size_gnsf_A[1]
    acados_ocp.dims.gnsf_nuhat = max(phi_fun.size_in(1))
    acados_ocp.dims.gnsf_ny = max(phi_fun.size_in(0))
    acados_ocp.dims.gnsf_nout = max(phi_fun.size_out(0))

    # # dim
    # model['dim_gnsf_nx1'] = gnsf['nx1']
    # model['dim_gnsf_nx2'] = gnsf['nx2']
    # model['dim_gnsf_nz1'] = gnsf['nz1']
    # model['dim_gnsf_nz2'] = gnsf['nz2']
    # model['dim_gnsf_nuhat'] = gnsf['nuhat']
    # model['dim_gnsf_ny'] = gnsf['ny']
    # model['dim_gnsf_nout'] = gnsf['n_out']

    # # sym
    # model['sym_gnsf_y'] = gnsf['y']
    # model['sym_gnsf_uhat'] = gnsf['uhat']

    # # data
    # model['dyn_gnsf_A'] = gnsf['A']
    # model['dyn_gnsf_A_LO'] = gnsf['A_LO']
    # model['dyn_gnsf_B'] = gnsf['B']
    # model['dyn_gnsf_B_LO'] = gnsf['B_LO']
    # model['dyn_gnsf_E'] = gnsf['E']
    # model['dyn_gnsf_E_LO'] = gnsf['E_LO']
    # model['dyn_gnsf_C'] = gnsf['C']
    # model['dyn_gnsf_c'] = gnsf['c']
    # model['dyn_gnsf_c_LO'] = gnsf['c_LO']
    # model['dyn_gnsf_L_x'] = gnsf['L_x']
    # model['dyn_gnsf_L_xdot'] = gnsf['L_xdot']
    # model['dyn_gnsf_L_z'] = gnsf['L_z']
    # model['dyn_gnsf_L_u'] = gnsf['L_u']
    # model['dyn_gnsf_idx_perm_x'] = gnsf['idx_perm_x']
    # model['dyn_gnsf_ipiv_x'] = gnsf['ipiv_x']
    # model['dyn_gnsf_idx_perm_z'] = gnsf['idx_perm_z']
    # model['dyn_gnsf_ipiv_z'] = gnsf['ipiv_z']
    # model['dyn_gnsf_idx_perm_f'] = gnsf['idx_perm_f']
    # model['dyn_gnsf_ipiv_f'] = gnsf['ipiv_f']

    # # flags
    acados_ocp.model.gnsf_nontrivial_f_LO = gnsf['nontrivial_f_LO']
    acados_ocp.model.gnsf_purely_linear = gnsf['purely_linear']

    # # casadi expr
    # model['dyn_gnsf_expr_phi'] = gnsf['phi_expr']
    # model['dyn_gnsf_expr_f_lo'] = gnsf['f_lo_expr']

    return acados_ocp



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

import casadi as ca

class GroupingTechnique:
    def __hash__(self):
        return hash(self.tuple)
    def __eq__(self, other):
        return self.tuple==other.tuple
    @property
    def tuple(self):
        return ()

    def __bool__(self):
        return False

class LseGroup(GroupingTechnique):
    class_id = 0
    def __init__(self, margin_abs=None):
        self.margin_abs = margin_abs

    @property
    def tuple(self):
        return (self.class_id, "margin_abs", self.margin_abs)

    def __call__(self, M, axis=0):
        if axis==1 and M.is_row():
            try:
                logsumexp = ca.logsumexp
            except:
                def logsumexp(x, margin):
                    alpha = ca.log(x.size1()) / margin
                    return ca.log(ca.sum1(ca.exp(alpha*x)))/alpha
            return logsumexp(M.T, self.margin_abs)
        else:
            raise Exception("Not implemented")

    def __bool__(self):
        return True


#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from typing import Union, List, Optional

import os
import casadi as ca
from .utils import is_empty, casadi_length, check_casadi_version_supports_p_global, print_casadi_expression
from .acados_model import AcadosModel
from .acados_ocp_constraints import AcadosOcpConstraints


def get_casadi_symbol(x):
    if isinstance(x, ca.MX):
        return ca.MX.sym
    elif isinstance(x, ca.SX):
        return ca.SX.sym
    else:
        raise TypeError("Expected casadi SX or MX.")

def is_casadi_SX(x):
    if isinstance(x, ca.SX):
        return True
    return False


class GenerateContext:
    def __init__(self, p_global: Optional[Union[ca.SX, ca.MX]], problem_name: str, opts=None):
        self.p_global = p_global
        if not is_empty(p_global):
            check_casadi_version_supports_p_global()

        self.problem_name = problem_name

        self.opts = opts
        self.casadi_codegen_opts = dict(mex=False, casadi_int='int', casadi_real='double')
        self.list_funname_dir_pairs = []  # list of (function_name, output_dir)

        self.generic_funname_dir_pairs = []  # list of (function_name, output_dir) of functions that are not generated by acados
        self.function_input_output_pairs: List[List[Union[ca.SX, ca.MX], Union[ca.SX, ca.MX]]] = []

        self.global_data_sym = None
        self.global_data_expr = None

        # check if CasADi version supports cse
        try:
            from casadi import cse
            casadi_fun_opts = {"cse": True}
        except:
            print("NOTE: Please consider updating to CasADi 3.6.6 which supports common subexpression elimination. \nThis might speed up external function evaluation.")
            casadi_fun_opts = {}

        self.__casadi_fun_opts = casadi_fun_opts


    def __generate_functions(self):
        for (name, output_dir), (inputs, outputs) in zip(self.list_funname_dir_pairs, self.function_input_output_pairs):
            # create function
            try:
                fun = ca.Function(name, inputs, outputs, self.__casadi_fun_opts)
                # print(f"Generating function {name} with inputs {inputs}")
            except Exception as e:
                print(f"\nError while creating function {name} with inputs {inputs} and outputs {outputs}")
                print(e)
                raise e

            # setup and change directory
            cwd = os.getcwd()
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
            os.chdir(output_dir)

            # generate function
            try:
                fun.generate(name, self.casadi_codegen_opts)
            except Exception as e:
                print(f"Error while generating function {name} in directory {output_dir}")
                print(e)
                raise e

            # change back to original directory
            os.chdir(cwd)

    def add_external_function_file(self, fun_name: str, output_dir: str):
        # remove trailing .c if present
        if fun_name.endswith(".c"):
            fun_name = fun_name[:-2]
        self.generic_funname_dir_pairs.append((fun_name, output_dir))

    def add_function_definition(self,
                                name: str,
                                inputs: List[Union[ca.MX, ca.SX]],
                                outputs: List[Union[ca.MX, ca.SX]],
                                output_dir: str):
        self.list_funname_dir_pairs.append((name, output_dir))
        self.function_input_output_pairs.append([inputs, outputs])

    def __setup_p_global_precompute_fun(self):
        precompute_pairs = []
        for i in range(len(self.function_input_output_pairs)):
            outputs = ca.cse(self.function_input_output_pairs[i][1])

            # detect parametric expressions in p_global
            [outputs_ret, symbols, param_expr] = ca.extract_parametric(outputs, self.p_global)

            # substitute previously detected param_expr in outputs
            symbols_to_add = []
            param_expr_to_add = []
            for sym_new, expr_new in zip(symbols, param_expr):
                add = True
                for sym, expr in precompute_pairs:
                    if ca.is_equal(expr, expr_new):
                        outputs_ret = [ca.substitute(output, sym_new, sym) for output in outputs_ret]
                        add = False
                        break
                if add:
                    symbols_to_add.append(sym_new)
                    param_expr_to_add.append(expr_new)

            # replace output expression with ones that use extracted expressions
            self.function_input_output_pairs[i][1] = outputs_ret
            # store (new input symbols, extracted expressions)
            for sym, expr in zip(symbols_to_add, param_expr_to_add):
                precompute_pairs.append([sym, expr])

        global_data_sym_list = [input for input, _ in precompute_pairs]
        self.global_data_sym = ca.vertcat(*global_data_sym_list)
        self.global_data_expr = ca.cse(ca.vertcat(*[output for _, output in precompute_pairs]))

        # add global data as input to all functions
        for i in range(len(self.function_input_output_pairs)):
            self.function_input_output_pairs[i][0].append(self.global_data_sym)

        output_dir = os.path.abspath(self.opts["code_export_directory"])
        fun_name = f'{self.problem_name}_p_global_precompute_fun'
        self.add_function_definition(fun_name, [self.p_global], [self.global_data_expr], output_dir)

        # self.print_global_data_summary()

        assert casadi_length(self.global_data_expr) == casadi_length(self.global_data_sym), f"Length mismatch: {casadi_length(self.global_data_expr)} != {casadi_length(self.global_data_sym)}"

    def finalize(self):
        if not is_empty(self.p_global):
            self.__setup_p_global_precompute_fun()

        self.__generate_functions()
        return

    def get_n_global_data(self):
        return casadi_length(self.global_data_sym)

    def get_external_function_file_list(self, ocp_specific=False):
        out = []
        for (fun_name, fun_dir) in self.generic_funname_dir_pairs + self.list_funname_dir_pairs:
            rel_fun_dir = os.path.relpath(fun_dir, self.opts["code_export_directory"])
            is_ocp_specific = not rel_fun_dir.endswith("model")
            if ocp_specific != is_ocp_specific:
                continue
            out.append(f"{rel_fun_dir}/{fun_name}.c")
        return out

    def print_global_data_summary(self) -> None:
        if casadi_length(self.global_data_expr) == 0:
            print("\nGenerateContext: detected empty global_data_expr.")
        else:
            print("\nGenerateContext: detected global_data_expr:\n")
            print_casadi_expression(self.global_data_expr)

################
# Dynamics
################

def generate_c_code_discrete_dynamics(context: GenerateContext, model: AcadosModel, model_dir: str):
    opts = context.opts

    # load model
    x = model.x
    u = model.u
    p = model.p
    p_global = model.p_global
    phi = model.disc_dyn_expr
    model_name = model.name

    symbol = get_casadi_symbol(x)
    nx1 = casadi_length(phi)

    lam = symbol('lam', nx1, 1)
    ux = ca.vertcat(u, x)

    # generate jacobians
    jac_ux = ca.jacobian(phi, ux)
    # generate adjoint
    adj_ux = ca.jtimes(phi, ux, lam, True)
    # generate hessian
    hess_ux = ca.jacobian(adj_ux, ux, {"symmetric": is_casadi_SX(x)})

    # set up & generate ca.Functions
    fun_name = model_name + '_dyn_disc_phi_fun'
    context.add_function_definition(fun_name, [x, u, p], [phi], model_dir)

    fun_name = model_name + '_dyn_disc_phi_fun_jac'
    context.add_function_definition(fun_name, [x, u, p], [phi, jac_ux.T], model_dir)

    fun_name = model_name + '_dyn_disc_phi_fun_jac_hess'
    context.add_function_definition(fun_name, [x, u, lam, p], [phi, jac_ux.T, hess_ux], model_dir)

    if opts["with_solution_sens_wrt_params"]:
        # generate jacobian of lagrange gradient wrt p
        jac_p = ca.jacobian(phi, p_global)
        # hess_xu_p_old = ca.jacobian((lam.T @ jac_ux).T, p)
        hess_xu_p = ca.jacobian(adj_ux, p_global) # using adjoint
        fun_name = model_name + '_dyn_disc_phi_jac_p_hess_xu_p'
        context.add_function_definition(fun_name, [x, u, lam, p], [jac_p, hess_xu_p], model_dir)

    if opts["with_value_sens_wrt_params"]:
        adj_p = ca.jtimes(phi, p_global, lam, True)
        fun_name = model_name + '_dyn_disc_phi_adj_p'
        context.add_function_definition(fun_name, [x, u, lam, p], [adj_p], model_dir)

    return



def generate_c_code_explicit_ode(context: GenerateContext, model: AcadosModel, model_dir: str):
    generate_hess = context.opts["generate_hess"]

    # load model
    x = model.x
    u = model.u
    p = model.p
    f_expl = model.f_expl_expr
    model_name = model.name

    nx = x.size()[0]
    nu = u.size()[0]

    symbol = get_casadi_symbol(x)

    # set up expressions
    Sx = symbol('Sx', nx, nx)
    Sp = symbol('Sp', nx, nu)
    lambdaX = symbol('lambdaX', nx, 1)

    vdeX = ca.jtimes(f_expl, x, Sx)
    vdeP = ca.jacobian(f_expl, u) + ca.jtimes(f_expl, x, Sp)
    adj = ca.jtimes(f_expl, ca.vertcat(x, u), lambdaX, True)

    if generate_hess:
        S_forw = ca.vertcat(ca.horzcat(Sx, Sp), ca.horzcat(ca.DM.zeros(nu,nx), ca.DM.eye(nu)))
        hess = ca.mtimes(ca.transpose(S_forw),ca.jtimes(adj, ca.vertcat(x,u), S_forw))
        hess2 = []
        for j in range(nx+nu):
            for i in range(j,nx+nu):
                hess2 = ca.vertcat(hess2, hess[i,j])

    # add to context
    fun_name = model_name + '_expl_ode_fun'
    context.add_function_definition(fun_name, [x, u, p], [f_expl], model_dir)

    fun_name = model_name + '_expl_vde_forw'
    context.add_function_definition(fun_name, [x, Sx, Sp, u, p], [f_expl, vdeX, vdeP], model_dir)

    fun_name = model_name + '_expl_vde_adj'
    context.add_function_definition(fun_name, [x, lambdaX, u, p], [adj], model_dir)

    if generate_hess:
        fun_name = model_name + '_expl_ode_hess'
        context.add_function_definition(fun_name, [x, Sx, Sp, lambdaX, u, p], [adj, hess2], model_dir)

    return


def generate_c_code_implicit_ode(context: GenerateContext, model: AcadosModel, model_dir: str):

    # load model
    x = model.x
    xdot = model.xdot
    u = model.u
    z = model.z
    p = model.p
    t = model.t
    f_impl = model.f_impl_expr
    model_name = model.name

    # get model dimensions
    nx = casadi_length(x)
    nz = casadi_length(z)

    # generate jacobians
    jac_x = ca.jacobian(f_impl, x)
    jac_xdot = ca.jacobian(f_impl, xdot)
    jac_u = ca.jacobian(f_impl, u)
    jac_z = ca.jacobian(f_impl, z)

    # Set up functions
    p = model.p
    fun_name = model_name + '_impl_dae_fun'
    context.add_function_definition(fun_name, [x, xdot, u, z, t, p], [f_impl], model_dir)

    fun_name = model_name + '_impl_dae_fun_jac_x_xdot_z'
    context.add_function_definition(fun_name, [x, xdot, u, z, t, p], [f_impl, jac_x, jac_xdot, jac_z], model_dir)

    fun_name = model_name + '_impl_dae_fun_jac_x_xdot_u_z'
    context.add_function_definition(fun_name, [x, xdot, u, z, t, p], [f_impl, jac_x, jac_xdot, jac_u, jac_z], model_dir)

    fun_name = model_name + '_impl_dae_fun_jac_x_xdot_u'
    context.add_function_definition(fun_name, [x, xdot, u, z, t, p], [f_impl, jac_x, jac_xdot, jac_u], model_dir)

    fun_name = model_name + '_impl_dae_jac_x_xdot_u_z'
    context.add_function_definition(fun_name, [x, xdot, u, z, t, p], [jac_x, jac_xdot, jac_u, jac_z], model_dir)

    if context.opts["generate_hess"]:
        x_xdot_z_u = ca.vertcat(x, xdot, z, u)
        symbol = get_casadi_symbol(x)
        multiplier = symbol('multiplier', nx + nz)
        ADJ = ca.jtimes(f_impl, x_xdot_z_u, multiplier, True)
        HESS = ca.jacobian(ADJ, x_xdot_z_u, {"symmetric": is_casadi_SX(x)})
        fun_name = model_name + '_impl_dae_hess'
        context.add_function_definition(fun_name, [x, xdot, u, z, multiplier, t, p], [HESS], model_dir)

    return


def generate_c_code_gnsf(context: GenerateContext, model: AcadosModel, model_dir: str):
    model_name = model.name

    # obtain gnsf dimensions
    get_matrices_fun = model.get_matrices_fun
    phi_fun = model.phi_fun

    size_gnsf_A = get_matrices_fun.size_out(0)
    gnsf_nx1 = size_gnsf_A[1]
    gnsf_nz1 = size_gnsf_A[0] - size_gnsf_A[1]
    gnsf_nuhat = max(phi_fun.size_in(1))
    gnsf_ny = max(phi_fun.size_in(0))

    # set up expressions
    # if the model uses ca.MX because of cost/constraints
    # the DAE can be exported as ca.SX -> detect GNSF in Matlab
    # -> evaluated ca.SX GNSF functions with ca.MX.
    u = model.u
    symbol = get_casadi_symbol(u)

    y = symbol("y", gnsf_ny, 1)
    uhat = symbol("uhat", gnsf_nuhat, 1)
    p = model.p
    x1 = symbol("gnsf_x1", gnsf_nx1, 1)
    x1dot = symbol("gnsf_x1dot", gnsf_nx1, 1)
    z1 = symbol("gnsf_z1", gnsf_nz1, 1)
    dummy = symbol("gnsf_dummy", 1, 1)
    empty_var = symbol("gnsf_empty_var", 0, 0)

    ## generate C code
    fun_name = model_name + '_gnsf_phi_fun'
    context.add_function_definition(fun_name, [y, uhat, p], [phi_fun(y, uhat, p)], model_dir)

    fun_name = model_name + '_gnsf_phi_fun_jac_y'
    phi_fun_jac_y = model.phi_fun_jac_y
    context.add_function_definition(fun_name, [y, uhat, p], phi_fun_jac_y(y, uhat, p), model_dir)

    fun_name = model_name + '_gnsf_phi_jac_y_uhat'
    phi_jac_y_uhat = model.phi_jac_y_uhat
    context.add_function_definition(fun_name, [y, uhat, p], phi_jac_y_uhat(y, uhat, p), model_dir)

    fun_name = model_name + '_gnsf_f_lo_fun_jac_x1k1uz'
    f_lo_fun_jac_x1k1uz = model.f_lo_fun_jac_x1k1uz
    f_lo_fun_jac_x1k1uz_eval = f_lo_fun_jac_x1k1uz(x1, x1dot, z1, u, p)

    # avoid codegeneration issue
    if not isinstance(f_lo_fun_jac_x1k1uz_eval, tuple) and is_empty(f_lo_fun_jac_x1k1uz_eval):
        f_lo_fun_jac_x1k1uz_eval = [empty_var]

    context.add_function_definition(fun_name, [x1, x1dot, z1, u, p], f_lo_fun_jac_x1k1uz_eval, model_dir)

    fun_name = model_name + '_gnsf_get_matrices_fun'
    context.add_function_definition(fun_name, [dummy], get_matrices_fun(1), model_dir)

    # remove fields for json dump
    del model.phi_fun
    del model.phi_fun_jac_y
    del model.phi_jac_y_uhat
    del model.f_lo_fun_jac_x1k1uz
    del model.get_matrices_fun

    return


################
# Cost
################

def generate_c_code_external_cost(context: GenerateContext, model: AcadosModel, stage_type):
    opts = context.opts

    x = model.x
    p = model.p
    u = model.u
    z = model.z
    p_global = model.p_global
    symbol = get_casadi_symbol(x)

    if stage_type == 'terminal':
        suffix_name = "_cost_ext_cost_e_fun"
        suffix_name_hess = "_cost_ext_cost_e_fun_jac_hess"
        suffix_name_jac = "_cost_ext_cost_e_fun_jac"
        suffix_name_param_sens = "_cost_ext_cost_e_hess_xu_p"
        suffix_name_value_sens = "_cost_ext_cost_e_grad_p"
        ext_cost = model.cost_expr_ext_cost_e
        custom_hess = model.cost_expr_ext_cost_custom_hess_e
        # Last stage cannot depend on u and z
        u = symbol("u", 0, 0)
        z = symbol("z", 0, 0)

    elif stage_type == 'path':
        suffix_name = "_cost_ext_cost_fun"
        suffix_name_hess = "_cost_ext_cost_fun_jac_hess"
        suffix_name_jac = "_cost_ext_cost_fun_jac"
        suffix_name_param_sens = "_cost_ext_cost_hess_xu_p"
        suffix_name_value_sens = "_cost_ext_cost_grad_p"
        ext_cost = model.cost_expr_ext_cost
        custom_hess = model.cost_expr_ext_cost_custom_hess

    elif stage_type == 'initial':
        suffix_name = "_cost_ext_cost_0_fun"
        suffix_name_hess = "_cost_ext_cost_0_fun_jac_hess"
        suffix_name_jac = "_cost_ext_cost_0_fun_jac"
        suffix_name_param_sens = "_cost_ext_cost_0_hess_xu_p"
        suffix_name_value_sens = "_cost_ext_cost_0_grad_p"
        ext_cost = model.cost_expr_ext_cost_0
        custom_hess = model.cost_expr_ext_cost_custom_hess_0

    nunx = x.shape[0] + u.shape[0]

    # set up functions to be exported
    fun_name = model.name + suffix_name
    fun_name_hess = model.name + suffix_name_hess
    fun_name_jac = model.name + suffix_name_jac
    fun_name_param = model.name + suffix_name_param_sens
    fun_name_value_sens = model.name + suffix_name_value_sens

    # generate expression for full gradient and Hessian
    hess_uxz, grad_uxz = ca.hessian(ext_cost, ca.vertcat(u, x, z))

    hess_ux = hess_uxz[:nunx, :nunx]
    hess_z = hess_uxz[nunx:, nunx:]
    hess_z_ux = hess_uxz[nunx:, :nunx]

    if not is_empty(custom_hess):
        hess_ux = custom_hess

    cost_dir = os.path.abspath(os.path.join(opts["code_export_directory"], f'{model.name}_cost'))

    context.add_function_definition(fun_name, [x, u, z, p], [ext_cost], cost_dir)
    context.add_function_definition(fun_name_hess, [x, u, z, p], [ext_cost, grad_uxz, hess_ux, hess_z, hess_z_ux], cost_dir)
    context.add_function_definition(fun_name_jac, [x, u, z, p], [ext_cost, grad_uxz], cost_dir)

    if opts["with_solution_sens_wrt_params"]:
        if casadi_length(z) > 0:
            raise Exception("acados: solution sensitivities wrt parameters not supported with algebraic variables.")
        grad_ux = ca.jacobian(ext_cost, ca.vertcat(u, x))
        hess_xu_p = ca.jacobian(grad_ux, p_global)
        context.add_function_definition(fun_name_param, [x, u, z, p], [hess_xu_p], cost_dir)

    if opts["with_value_sens_wrt_params"]:
        grad_p = ca.jacobian(ext_cost, p_global).T
        context.add_function_definition(fun_name_value_sens, [x, u, z, p], [grad_p], cost_dir)

    return


def generate_c_code_nls_cost(context: GenerateContext, model: AcadosModel, stage_type):
    opts = context.opts

    x = model.x
    z = model.z
    p = model.p
    u = model.u
    t = model.t

    symbol = get_casadi_symbol(x)

    if stage_type == 'terminal':
        middle_name = '_cost_y_e'
        u = symbol('u', 0, 0)
        y_expr = model.cost_y_expr_e

    elif stage_type == 'initial':
        middle_name = '_cost_y_0'
        y_expr = model.cost_y_expr_0

    elif stage_type == 'path':
        middle_name = '_cost_y'
        y_expr = model.cost_y_expr

    cost_dir = os.path.abspath(os.path.join(opts["code_export_directory"], f'{model.name}_cost'))

    # set up expressions
    cost_jac_expr = ca.transpose(ca.jacobian(y_expr, ca.vertcat(u, x)))
    dy_dz = ca.jacobian(y_expr, z)
    ny = casadi_length(y_expr)

    # Check if dimension is 0, otherwise Casadi will crash
    y = symbol('y', ny, 1)
    if ny == 0:
        y_adj = 0
        y_hess = 0
    else:
        y_adj = ca.jtimes(y_expr, ca.vertcat(u, x), y, True)
        y_hess = ca.jacobian(y_adj, ca.vertcat(u, x), {"symmetric": is_casadi_SX(x)})

    ## generate C code
    suffix_name = '_fun'
    fun_name = model.name + middle_name + suffix_name
    context.add_function_definition(fun_name, [x, u, z, t, p], [ y_expr ], cost_dir)

    suffix_name = '_fun_jac_ut_xt'
    fun_name = model.name + middle_name + suffix_name
    context.add_function_definition(fun_name, [x, u, z, t, p], [ y_expr, cost_jac_expr, dy_dz ], cost_dir)

    suffix_name = '_hess'
    fun_name = model.name + middle_name + suffix_name
    context.add_function_definition(fun_name, [x, u, z, y, t, p], [ y_hess ], cost_dir)

    return



def generate_c_code_conl_cost(context: GenerateContext, model: AcadosModel, stage_type: str):

    opts = context.opts
    x = model.x
    z = model.z
    p = model.p
    p_global = model.p_global
    t = model.t

    symbol = get_casadi_symbol(x)
    p_global = symbol('p_global', 0, 0)

    if stage_type == 'terminal':
        u = symbol('u', 0, 0)

        yref = model.cost_r_in_psi_expr_e
        inner_expr = model.cost_y_expr_e - yref
        outer_expr = model.cost_psi_expr_e
        res_expr = model.cost_r_in_psi_expr_e

        suffix_name_fun = '_conl_cost_e_fun'
        suffix_name_fun_jac_hess = '_conl_cost_e_fun_jac_hess'

        custom_hess = model.cost_conl_custom_outer_hess_e

    elif stage_type == 'initial':
        u = model.u

        yref = model.cost_r_in_psi_expr_0
        inner_expr = model.cost_y_expr_0 - yref
        outer_expr = model.cost_psi_expr_0
        res_expr = model.cost_r_in_psi_expr_0

        suffix_name_fun = '_conl_cost_0_fun'
        suffix_name_fun_jac_hess = '_conl_cost_0_fun_jac_hess'

        custom_hess = model.cost_conl_custom_outer_hess_0

    elif stage_type == 'path':
        u = model.u

        yref = model.cost_r_in_psi_expr
        inner_expr = model.cost_y_expr - yref
        outer_expr = model.cost_psi_expr
        res_expr = model.cost_r_in_psi_expr

        suffix_name_fun = '_conl_cost_fun'
        suffix_name_fun_jac_hess = '_conl_cost_fun_jac_hess'

        custom_hess = model.cost_conl_custom_outer_hess

    # set up function names
    fun_name_cost_fun = model.name + suffix_name_fun
    fun_name_cost_fun_jac_hess = model.name + suffix_name_fun_jac_hess

    # set up functions to be exported
    outer_loss_fun = ca.Function('psi', [res_expr, t, p, p_global], [outer_expr])
    cost_expr = outer_loss_fun(inner_expr, t, p, p_global)

    outer_loss_grad_fun = ca.Function('outer_loss_grad', [res_expr, t, p, p_global], [ca.jacobian(outer_expr, res_expr).T])

    if is_empty(custom_hess):
        hess = ca.hessian(outer_loss_fun(res_expr, t, p, p_global), res_expr)[0]
    else:
        hess = custom_hess

    outer_hess_fun = ca.Function('outer_hess', [res_expr, t, p, p_global], [hess])
    outer_hess_expr = outer_hess_fun(inner_expr, t, p, p_global)
    outer_hess_is_diag = outer_hess_expr.sparsity().is_diag()
    if casadi_length(res_expr) <= 4:
        outer_hess_is_diag = 0

    Jt_ux_expr = ca.jacobian(inner_expr, ca.vertcat(u, x)).T
    Jt_z_expr = ca.jacobian(inner_expr, z).T

    # change directory
    cost_dir = os.path.abspath(os.path.join(opts["code_export_directory"], f'{model.name}_cost'))

    context.add_function_definition(
        fun_name_cost_fun,
        [x, u, z, yref, t, p],
        [cost_expr], cost_dir)

    context.add_function_definition(
        fun_name_cost_fun_jac_hess,
        [x, u, z, yref, t, p],
        [cost_expr, outer_loss_grad_fun(inner_expr, t, p, p_global), Jt_ux_expr, Jt_z_expr, outer_hess_expr, outer_hess_is_diag],
        cost_dir
    )

    return


################
# Constraints
################
def generate_c_code_constraint(context: GenerateContext, model: AcadosModel, constraints: AcadosOcpConstraints, stage_type: str):

    opts = context.opts

    # load constraint variables and expression
    x = model.x
    p = model.p
    u = model.u
    z = model.z

    symbol = get_casadi_symbol(x)

    if stage_type == 'terminal':
        constr_type = constraints.constr_type_e
        con_h_expr = model.con_h_expr_e
        con_phi_expr = model.con_phi_expr_e
        # create dummy u, z
        u = symbol('u', 0, 0)
        z = symbol('z', 0, 0)
    elif stage_type == 'initial':
        constr_type = constraints.constr_type_0
        con_h_expr = model.con_h_expr_0
        con_phi_expr = model.con_phi_expr_0
    elif stage_type == 'path':
        constr_type = constraints.constr_type
        con_h_expr = model.con_h_expr
        con_phi_expr = model.con_phi_expr

    if (not is_empty(con_h_expr)) and (not is_empty(con_phi_expr)):
        raise Exception("acados: you can either have constraint_h, or constraint_phi, not both.")

    if (is_empty(con_h_expr) and is_empty(con_phi_expr)):
        # both empty -> nothing to generate
        return

    if is_empty(p):
        p = symbol('p', 0, 0)

    if is_empty(z):
        z = symbol('z', 0, 0)

    # multipliers for hessian
    nh = casadi_length(con_h_expr)
    lam_h = symbol('lam_h', nh, 1)

    # directory
    constraints_dir = os.path.abspath(os.path.join(opts["code_export_directory"], f'{model.name}_constraints'))

    # export casadi functions
    if constr_type == 'BGH':
        if stage_type == 'terminal':
            fun_name = model.name + '_constr_h_e_fun_jac_uxt_zt'
        elif stage_type == 'initial':
            fun_name = model.name + '_constr_h_0_fun_jac_uxt_zt'
        else:
            fun_name = model.name + '_constr_h_fun_jac_uxt_zt'

        jac_ux_t = ca.transpose(ca.jacobian(con_h_expr, ca.vertcat(u,x)))
        jac_z_t = ca.jacobian(con_h_expr, z)
        context.add_function_definition(fun_name, [x, u, z, p], \
                [con_h_expr, jac_ux_t, jac_z_t], constraints_dir)

        if opts['generate_hess']:
            if stage_type == 'terminal':
                fun_name = model.name + '_constr_h_e_fun_jac_uxt_zt_hess'
            elif stage_type == 'initial':
                fun_name = model.name + '_constr_h_0_fun_jac_uxt_zt_hess'
            else:
                fun_name = model.name + '_constr_h_fun_jac_uxt_zt_hess'

            # adjoint
            adj_ux = ca.jtimes(con_h_expr, ca.vertcat(u, x), lam_h, True)
            # hessian
            hess_ux = ca.jacobian(adj_ux, ca.vertcat(u, x), {"symmetric": is_casadi_SX(x)})

            adj_z = ca.jtimes(con_h_expr, z, lam_h, True)
            hess_z = ca.jacobian(adj_z, z, {"symmetric": is_casadi_SX(x)})

            context.add_function_definition(fun_name, [x, u, lam_h, z, p], \
                    [con_h_expr, jac_ux_t, hess_ux, jac_z_t, hess_z], constraints_dir)

        if stage_type == 'terminal':
            fun_name = model.name + '_constr_h_e_fun'
        elif stage_type == 'initial':
            fun_name = model.name + '_constr_h_0_fun'
        else:
            fun_name = model.name + '_constr_h_fun'
        context.add_function_definition(fun_name, [x, u, z, p], [con_h_expr], constraints_dir)

        if opts["with_solution_sens_wrt_params"]:
            jac_p = ca.jacobian(con_h_expr, model.p_global)
            adj_ux = ca.jtimes(con_h_expr, ca.vertcat(u, x), lam_h, True)
            hess_xu_p = ca.jacobian(adj_ux, model.p_global)

            if stage_type == 'terminal':
                fun_name = model.name + '_constr_h_e_jac_p_hess_xu_p'
            elif stage_type == 'initial':
                fun_name = model.name + '_constr_h_0_jac_p_hess_xu_p'
            else:
                fun_name = model.name + '_constr_h_jac_p_hess_xu_p'

            context.add_function_definition(fun_name, [x, u, lam_h, z, p], \
                    [jac_p, hess_xu_p], constraints_dir)

        if opts["with_value_sens_wrt_params"]:
            adj_p = ca.jtimes(con_h_expr, model.p_global, lam_h, True)
            if stage_type == 'terminal':
                fun_name = model.name + '_constr_h_e_adj_p'
            elif stage_type == 'initial':
                fun_name = model.name + '_constr_h_0_adj_p'
            else:
                fun_name = model.name + '_constr_h_adj_p'

            context.add_function_definition(fun_name, [x, u, lam_h, p], [adj_p], constraints_dir)

    else: # BGP constraint
        if stage_type == 'terminal':
            fun_name_prefix = model.name + '_phi_e_constraint'
            r = model.con_r_in_phi_e
            con_r_expr = model.con_r_expr_e
        elif stage_type == 'initial':
            fun_name_prefix = model.name + '_phi_0_constraint'
            r = model.con_r_in_phi_0
            con_r_expr = model.con_r_expr_0
        elif stage_type == 'path':
            fun_name_prefix = model.name + '_phi_constraint'
            r = model.con_r_in_phi
            con_r_expr = model.con_r_expr

        nphi = casadi_length(con_phi_expr)
        con_phi_expr_x_u_z = ca.substitute(con_phi_expr, r, con_r_expr)
        phi_jac_u = ca.jacobian(con_phi_expr_x_u_z, u)
        phi_jac_x = ca.jacobian(con_phi_expr_x_u_z, x)
        phi_jac_z = ca.jacobian(con_phi_expr_x_u_z, z)

        hess = ca.vertcat(*[ca.hessian(con_phi_expr[i], r)[0] for i in range(nphi)])
        hess = ca.substitute(hess, r, con_r_expr)

        r_jac_u = ca.jacobian(con_r_expr, u)
        r_jac_x = ca.jacobian(con_r_expr, x)

        fun_jac_hess_name = fun_name_prefix + '_fun_jac_hess'
        context.add_function_definition(fun_jac_hess_name, [x, u, z, p], \
                [con_phi_expr_x_u_z, \
                ca.vertcat(ca.transpose(phi_jac_u), ca.transpose(phi_jac_x)), \
                ca.transpose(phi_jac_z), \
                hess,
                ca.vertcat(ca.transpose(r_jac_u), ca.transpose(r_jac_x))],
                constraints_dir
                )

        fun_name = fun_name_prefix + '_fun'
        context.add_function_definition(fun_name, [x, u, z, p], [con_phi_expr_x_u_z], constraints_dir)

    return




from rockit import *
      
mode = "nominal"

ocp = Ocp(T=1.0)

x      = ocp.state()
v      = ocp.state()

F       = ocp.control()

ocp.set_der(x, v)
ocp.set_der(v, F)

ocp.add_objective(ocp.sum(v**2))
ocp.add_objective(5*ocp.at_tf(v)**2)

ocp.subject_to(ocp.at_t0(x)==0.0)
ocp.subject_to(ocp.at_t0(v)==0.0)

constraints = []


constr_type = "nonlinear"
if constr_type=="simple":
    constraints.append((lambda e: F <= 15+e,True,{})) # simple
    constraints.append((lambda e: v <= 7+e,True,dict(include_first=False,include_last=False)))  # include_first=True induces failure
    constraints.append((lambda e: ocp.at_tf(x) + e>=5.0,False,{})) # boundary constraint on state
elif constr_type=="linear":
    constraints.append((lambda e: F <= 15+e,True,{})) # simple
    constraints.append((lambda e: v+0.01*x <= 7+e,True,dict(include_first=True,include_last=False)))  # include_first=True induces failure
    constraints.append((lambda e: ocp.at_tf(x+0.01*v) + e>=5.0,False,{})) # boundary constraint on state
elif constr_type=="nonlinear":
    constraints.append((lambda e: F <= 15+e,True,{})) # simple
    constraints.append((lambda e: v+0.01*x**2 <= 7+e,True,dict(include_first=False,include_last=False)))  # include_first=True induces failure
    constraints.append((lambda e: ocp.at_tf(x+0.01*v**2) + e>=5.0,False,{}))



if mode=="nominal":
    for c,_,kwargs in constraints:
        ocp.subject_to(c(0),**kwargs)
elif mode=="perturbed":
    k = 0
    for i,(c,_,kwargs) in enumerate(constraints):
        ocp.subject_to(c(0.1 if i==k else 0),**kwargs)
elif mode=="soft":
    k = 0
    power = 1 # 1-norm or 2-norm?
    for i,(c,signal,kwargs) in enumerate(constraints):
        if i!=k:
            ocp.subject_to(c(0),**kwargs)
    else:
        slack = ocp.variable(grid='control' if signal else '')
        ocp.subject_to(slack>=0)
        ocp.subject_to(c(slack),**kwargs)
        penalty = 0.3*slack**power
        print("signal",signal, penalty)
        ocp.add_objective(ocp.sum(penalty) if signal else penalty)

ocp.method(MultipleShooting(N=4,intg='rk'))

ocp.solver('ipopt',{"ipopt.tol":1e-10})

sol = ocp.solve()


signals = [("x",x),("v",v),("F",F)]
values = []#("mayer",mayer)]#[("slack",slack)]

ref = {}
for k, expr in signals:
    ref[k] = sol.sample(expr,grid='control')[1]
for k, expr in values:
    ref[k] = sol.value(expr)

[ref_t,_] = sol.sample(signals[0][1],grid='control')

print(ref)

#raise Exception("")

for method in [external_method('acados',N=4,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=2000,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=4)]:

    ocp.method(method)
    sol = ocp.solve()

    sold = {}
    for k, expr in signals:
        sold[k] = sol.sample(expr,grid='control')[1]
    for k, expr in values:
        sold[k] = sol.value(expr)

    [sol_t,_] = sol.sample(signals[0][1],grid='control')

    tolerance = 1e-5

    for k, expr in signals:
        print(k+"ref",ref[k])
        print(k+"sol",sold[k])
    for k, expr in values:
        print(k+"ref",ref[k])
        print(k+"sol",sold[k])
        print("ref_t",ref_t)
        print("sol_t",sol_t)



###################################################################################################
#                                                                                                 #
# This file is part of HPIPM.                                                                     #
#                                                                                                 #
# HPIPM -- High-Performance Interior Point Method.                                                #
# Copyright (C) 2019 by Gianluca Frison.                                                          #
# Developed at IMTEK (University of Freiburg) under the supervision of Moritz Diehl.              #
# All rights reserved.                                                                            #
#                                                                                                 #
# The 2-Clause BSD License                                                                        #
#                                                                                                 #
# Redistribution and use in source and binary forms, with or without                              #
# modification, are permitted provided that the following conditions are met:                     #
#                                                                                                 #
# 1. Redistributions of source code must retain the above copyright notice, this                  #
#    list of conditions and the following disclaimer.                                             #
# 2. Redistributions in binary form must reproduce the above copyright notice,                    #
#    this list of conditions and the following disclaimer in the documentation                    #
#    and/or other materials provided with the distribution.                                       #
#                                                                                                 #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND                 #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED                   #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE                          #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR                 #
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES                  #
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;                    #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND                     #
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT                      #
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS                   #
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                    #
#                                                                                                 #
# Author: Gianluca Frison, gianluca.frison (at) imtek.uni-freiburg.de                             #
#                                                                                                 #
###################################################################################################

from ctypes import *
import ctypes.util
import numpy as np



class hpipm_ocp_qcqp_sol:


	def __init__(self, dim):

		# save dim internally
		self.dim = dim

		# load hpipm shared library
		__hpipm   = CDLL('libhpipm.so')
		self.__hpipm = __hpipm

		# C qp struct
		qp_sol_struct_size = __hpipm.d_ocp_qcqp_sol_strsize()
		qp_sol_struct = cast(create_string_buffer(qp_sol_struct_size), c_void_p)
		self.qp_sol_struct = qp_sol_struct

		# C qp internal memory
		qp_sol_mem_size = __hpipm.d_ocp_qcqp_sol_memsize(dim.dim_struct)
		qp_sol_mem = cast(create_string_buffer(qp_sol_mem_size), c_void_p)
		self.qp_sol_mem = qp_sol_mem

		# create C qp
		__hpipm.d_ocp_qcqp_sol_create(dim.dim_struct, qp_sol_struct, qp_sol_mem)


	def get(self, field, idx_start, idx_end=None):
		if(field=='u'):
			vec = self.__get_u(idx_start, idx_end)
		elif(field=='x'):
			vec = self.__get_x(idx_start, idx_end)
		else:
			raise NameError('hpipm_ocp_qcqp_sol.get: wrong field')
		return vec


	def __get_u(self, idx_start, idx_end=None):
		# nu
		nu = np.zeros((1,1), dtype=int);
		if idx_end==None:
			# get nu at stage
			tmp_ptr = cast(nu.ctypes.data, POINTER(c_int))
			self.__hpipm.d_ocp_qcqp_dim_get_nu(self.dim.dim_struct, idx_start, tmp_ptr)
			u = np.zeros((nu[0,0], 1))
			tmp_ptr = cast(u.ctypes.data, POINTER(c_double))
			self.__hpipm.d_ocp_qcqp_sol_get_u(idx_start, self.qp_sol_struct, tmp_ptr)
		else:
			u = []
			for i in range(idx_start, idx_end+1):
				# get nu at stage
				tmp_ptr = cast(nu.ctypes.data, POINTER(c_int))
				self.__hpipm.d_ocp_qcqp_dim_get_nu(self.dim.dim_struct, i, tmp_ptr)
				u.append(np.zeros((nu[0,0], 1)))
				tmp_ptr = cast(u[-1].ctypes.data, POINTER(c_double))
				self.__hpipm.d_ocp_qcqp_sol_get_u(i, self.qp_sol_struct, tmp_ptr)
		return u


	def __get_x(self, idx_start, idx_end=None):
		# nx
		nx = np.zeros((1,1), dtype=int);
		if idx_end==None:
			# get nx at stage
			tmp_ptr = cast(nx.ctypes.data, POINTER(c_int))
			self.__hpipm.d_ocp_qcqp_dim_get_nx(self.dim.dim_struct, idx_start, tmp_ptr)
			x = np.zeros((nx[0,0], 1))
			tmp_ptr = cast(x.ctypes.data, POINTER(c_double))
			self.__hpipm.d_ocp_qcqp_sol_get_x(idx_start, self.qp_sol_struct, tmp_ptr)
		else:
			x = []
			for i in range(idx_start, idx_end+1):
				# get nx at stage
				tmp_ptr = cast(nx.ctypes.data, POINTER(c_int))
				self.__hpipm.d_ocp_qcqp_dim_get_nx(self.dim.dim_struct, i, tmp_ptr)
				x.append(np.zeros((nx[0,0], 1)))
				tmp_ptr = cast(x[-1].ctypes.data, POINTER(c_double))
				self.__hpipm.d_ocp_qcqp_sol_get_x(i, self.qp_sol_struct, tmp_ptr)
		return x


	def print_C_struct(self):
		self.__hpipm.d_ocp_qcqp_sol_print(self.dim.dim_struct, self.qp_sol_struct)
		return



import unittest

from pylab import *

from rockit import Ocp, DirectMethod, MultipleShooting, DirectCollocation, SingleShooting, SplineMethod
from problems import integrator_control_problem, bang_bang_chain_problem
import numpy as np

class MethodTests(unittest.TestCase):

    def test_all(self):
      T = 1
      M = 1
      b = 1
      t0 = 0
      x0 = 0
      for scheme in [MultipleShooting(N=40, M=1, intg='rk'),
                     DirectCollocation(N=40), SingleShooting(N=40, M=1, intg='rk'),
                     SplineMethod(N=10)]:
        (ocp, x, u) = integrator_control_problem(T, b, x0, scheme, t0)
        sol = ocp.solve()
        ts, xs = sol.sample(x, grid='control')

        self.assertAlmostEqual(xs[0],x0,places=6)
        self.assertAlmostEqual(xs[-1],x0-b*T,places=6)
  

    def test_spline_more_vars(self):
      (ocp, p, v, a) = bang_bang_chain_problem(MultipleShooting(N=10, intg='rk'))
      sol = ocp.solve()
      ts, ref_ps = sol.sample(p, grid='control')
      ts, ref_vs = sol.sample(v, grid='control')
      ts, ref_as = sol.sample(a, grid='control')
      for scheme in [DirectCollocation(N=10), SingleShooting(N=10, intg='rk'), SplineMethod(N=10)]:
        (ocp, p, v, a) = bang_bang_chain_problem(scheme)
        sol = ocp.solve()
        ts, ps = sol.sample(p, grid='control')
        ts, vs = sol.sample(v, grid='control')
        ts, ass = sol.sample(a, grid='control')

        np.testing.assert_allclose(ps, ref_ps, atol=1e-6)
        np.testing.assert_allclose(vs, ref_vs, atol=1e-6)
        np.testing.assert_allclose(ass, ref_as, atol=1e-6)
        

        
if __name__ == '__main__':
    unittest.main()



# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from typing import Optional, Tuple
import casadi as ca

def huber_loss(var: ca.SX, delta: float, tau: float) -> Tuple[ca.SX, ca.SX, ca.SX, ca.SX]:
    """
    Huber loss function.
    Return the loss, gradient, hessian and the XGN Hessian
    """
    loss = (tau / delta) * ca.if_else(
        ca.fabs(var) < delta, 0.5 * var**2, delta * (ca.fabs(var) - 0.5 * delta)
    )

    loss_hess, loss_grad = ca.hessian(loss, var)
    loss_hess_XGN = ca.if_else(var == 0, loss_hess, ca.diag(loss_grad / var))

    return loss, loss_grad, loss_hess, loss_hess_XGN


def one_sided_huber_penalty(
    u: ca.SX,
    delta: float,
    tau: Optional[float] = None,
    w: Optional[float] = None,
    min_hess: float = 0.0,
):
    """
    One-sided Huber penalty for a constraint u <= 0.
    Note: either tau or w need to be specified.
    delta: the length of the quadratic behavior
    tau: gradient in linear region
    w: hessian in quadratic region
    min_hess: provide a minimum value for the hessian
    """

    if delta < 0:
        raise ValueError("delta must be positive")

    if tau is None:
        if w is None:
            raise Exception("Either specify w or tau")
        tau = 2 * w * delta
    elif w is not None:
        raise Exception("Either specify w or tau")

    loss, _, _, loss_hess_XGN = huber_loss(u, delta, tau)
    # shifted by delta to get a penalty
    penalty = 0.5 * (ca.substitute(loss, u, u - delta) + tau * u)

    penalty_0 = ca.substitute(penalty, u, 0)
    penalty = penalty - penalty_0

    penalty_hess, penalty_grad = ca.hessian(penalty, u)

    penalty_hess_xgn = 0.5 * ca.substitute(loss_hess_XGN, u, u - delta)

    if min_hess > 0.0:
        penalty_hess = ca.fmax(min_hess, penalty_hess)
        penalty_hess_xgn = ca.fmax(min_hess, penalty_hess_xgn)

    return penalty, penalty_grad, penalty_hess, penalty_hess_xgn


def symmetric_huber_penalty(
    u: ca.SX,
    delta: float,
    tau: Optional[float] = None,
    w: Optional[float] = None,
    min_hess: float = 0.0,
):
    """
    Symmetric Huber penalty for a constraint -1 <= u <= 1.
    Note: either tau or w need to be specified.
    delta: the length of the quadratic behavior
    tau: gradient in linear region
    w: hessian in quadratic region
    min_hess: provide a minimum value for the hessian
    """

    if delta < 0:
        raise ValueError("delta must be positive")

    if tau is None:
        if w is None:
            raise Exception("Either specify w or tau")
        tau = 2 * w * delta
    elif w is not None:
        raise Exception("Either specify w or tau")

    loss, _, _, loss_hess_XGN = huber_loss(u, delta, tau)

    # barrier huber
    # penalty = 0.5*(ca.substitute(loss, u, u - 1) + ca.substitute(loss, u, u + 1) - ca.substitute(loss, u, -1) - ca.substitute(loss, u, 1))

    # shifted by delta to get a penalty
    penalty = 0.5 * (
        ca.substitute(loss, u, u - (1 + delta))
        + ca.substitute(loss, u, u + (1 + delta))
    )
    penalty += 0.5 * (
        -ca.substitute(loss, u, -(1 + delta)) - ca.substitute(loss, u, 1 - delta)
    )

    penalty_0 = ca.substitute(penalty, u, 0)
    penalty = penalty - penalty_0

    penalty_hess, penalty_grad = ca.hessian(penalty, u)

    penalty_hess_xgn = 0.5 * ca.if_else(
        u < 0,
        ca.substitute(loss_hess_XGN, u, u + 1 + delta),
        ca.substitute(loss_hess_XGN, u, u - 1 - delta),
    )

    if min_hess > 0.0:
        penalty_hess = ca.fmax(min_hess, penalty_hess)
        penalty_hess_xgn = ca.fmax(min_hess, penalty_hess_xgn)

    return penalty, penalty_grad, penalty_hess, penalty_hess_xgn




# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;


from dataclasses import dataclass
from typing import List
import numpy as np


@dataclass
class AcadosOcpFlattenedIterate:
    x: np.ndarray
    u: np.ndarray
    z: np.ndarray
    sl: np.ndarray
    su: np.ndarray
    pi: np.ndarray
    lam: np.ndarray


@dataclass
class AcadosOcpFlattenedBatchIterate:
    x: np.ndarray # shape (N_batch, nx_total)
    u: np.ndarray # shape (N_batch, nu_total)
    z: np.ndarray
    sl: np.ndarray
    su: np.ndarray
    pi: np.ndarray
    lam: np.ndarray
    N_batch: int


@dataclass
class AcadosOcpIterate:

    x_traj: List[np.ndarray]
    u_traj: List[np.ndarray]
    z_traj: List[np.ndarray]
    sl_traj: List[np.ndarray]
    su_traj: List[np.ndarray]
    pi_traj: List[np.ndarray]
    lam_traj: List[np.ndarray]

    def flatten(self) -> AcadosOcpFlattenedIterate:
        return AcadosOcpFlattenedIterate(
            x=np.concatenate(self.x_traj),
            u=np.concatenate(self.u_traj),
            z=np.concatenate(self.z_traj),
            sl=np.concatenate(self.sl_traj),
            su=np.concatenate(self.su_traj),
            pi=np.concatenate(self.pi_traj),
            lam=np.concatenate(self.lam_traj),
        )

@dataclass
class AcadosOcpIterates:

    iterate_list: List[AcadosOcpIterate]
    __iterate_fields = ["x", "u", "z", "sl", "su", "pi", "lam"]

    def as_array(self, field: str, ) -> np.ndarray:
        """
        Returns the iterate given by field as a numpy array of shape (nlp_iter+1, N_horizon (+ 1), n_field).
        This will fail if the dimension of value `field` is varying stagewise.
        """

        if field not in self.__iterate_fields:
            raise Exception(f"Invalid field: got {field}, expected value in {self.__iterate_fields}")

        attr = f"{field}_traj"
        traj_ = [getattr(iterate, attr) for iterate in self.iterate_list]

        try:
            traj = np.array(traj_, dtype=float)
        except ValueError:
            raise Exception(f"Stage-wise dimensions are not the same for {field} trajectory.")

        return traj
